@@ -371,5 +371,5 @@ SEXP initLastUpdated(SEXP var) {
 
 SEXP dllVersion(void) {
   // .onLoad calls this and checks the same as packageVersion() to ensure no R/C version mismatch, #3056
-  return(ScalarString(mkChar("1.18.0")));
+  return(ScalarString(mkChar("1.18.2.1")));
 }
