@@ -1,3 +1,11 @@
+
+// Compatibility define for LLVM builds of libomp.
+#ifdef KMP_VERSION_BUILD
+# ifndef _OPENMP
+#  define _OPENMP KMP_VERSION_BUILD
+# endif
+#endif
+
 #ifdef _OPENMP
   #include <omp.h>
   #if _OPENMP >= 201511
@@ -15,8 +23,6 @@
   #define omp_get_max_threads()  1
   #define omp_get_thread_limit() 1
   #define omp_get_num_procs()    1
-  #define omp_set_nested(a)   // empty statement to remove the call
   #define omp_get_wtime()        0
   #define MY_OPENMP              0
 #endif
-
